"""Login functionality for the LTD Keeper API.
"""

__all__ = ('get_keeper_token',)

from urllib.parse import urljoin
import requests

from .exceptions import KeeperError


def get_keeper_token(host, username, password):
    """Get a temporary auth token from LTD Keeper.

    Parameters
    ----------
    host : `str`
        Hostname of the LTD Keeper API (e.g., ``'https://keeper.lsst.codes'``).
    username : `str`
        Username.
    password : `str`
        Password.

    Returns
    -------
    token : `str`
        LTD Keeper API token.

    Raises
    ------
    KeeperError
        Raised if the LTD Keeper API cannot return a token.
    """
    token_endpoint = urljoin(host, '/token')
    r = requests.get(token_endpoint, auth=(username, password))
    if r.status_code != 200:
        raise KeeperError('Could not authenticate to {0}: error {1:d}\n{2}'.
                          format(host, r.status_code, r.json()))
    return r.json()['token']



"""Exceptions related to the LTD Keeper.
"""

__all__ = ('KeeperError',)

from ..exceptions import ConveyorError


class KeeperError(ConveyorError):
    """Error raised because of issues using the LTD Keeper API.
    """



import os
import sys

from documenteer.sphinxconfig.utils import form_ltd_edition_name
import lsst_sphinx_bootstrap_theme


# Work around Sphinx bug related to large and highly-nested source files
sys.setrecursionlimit(2000)

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.ifconfig',
    'sphinx_click.ext',
    'numpydoc',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
    'documenteer.sphinxext'
]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'LTD Conveyor'
copyright = '2017-2020 Association of Universities for Research in Astronomy, Inc. (AURA)'
author = 'LSST Data Management'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
if os.getenv('TRAVIS_BRANCH', default='master') == 'master':
    # Use the current release as the version tag if on master
    version = 'Current'
    release = version
else:
    # Use branch name as the version tag
    version = form_ltd_edition_name(
        git_ref_name=os.getenv('TRAVIS_BRANCH', default='master'))
    release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [
    '_build',
    'README.rst'
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# The reST default role cross-links Python (used for this markup: `text`)
default_role = 'py:obj'

# Intersphinx

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'boto3': ('http://boto3.readthedocs.io/en/latest/', None),
}

# -- Options for linkcheck builder ----------------------------------------

linkcheck_retries = 2

# -- Options for HTML output ----------------------------------------------

templates_path = [
    '_templates',
    lsst_sphinx_bootstrap_theme.get_html_templates_path()
]

html_theme = 'lsst_sphinx_bootstrap_theme'
html_theme_path = [lsst_sphinx_bootstrap_theme.get_html_theme_path()]


html_context = {
    # Enable "Edit in GitHub" link
    'display_github': True,
    # https://{{ github_host|default("github.com") }}/{{ github_user }}/
    #     {{ github_repo }}/blob/
    #     {{ github_version }}{{ conf_py_path }}{{ pagename }}{{ suffix }}
    'github_user': 'lsst-sqre',
    'github_repo': 'documenteer',
    'conf_py_path': 'docs/',
    # TRAVIS_BRANCH is available in CI, but master is a safe default
    'github_version': os.getenv('TRAVIS_BRANCH', default='master') + '/'
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {'logotext': project}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = project

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False



"""Tests for s3delete.

These tests required the following environment variables:

``LTD_TEST_AWS_ID``
    AWS access key ID

``LTD_TEST_AWS_SECRET``
    AWS secret access key

``LTD_TEST_BUCKET``
    Name of an S3 bucket that already exists and can be used for testing.

The tests will be skipped if they are not available.

Note that this test will create a random uuid4) directory at the root of
``LTD_TEST_BUCKET``, though the test harness will attempt to delete
it.
"""

import os
import uuid

import boto3
import pytest

from ltdconveyor.s3 import delete_dir
from ltdconveyor.testutils import upload_test_files


@pytest.mark.skipif(os.getenv('LTD_TEST_AWS_ID') is None or
                    os.getenv('LTD_TEST_AWS_SECRET') is None or
                    os.getenv('LTD_TEST_BUCKET') is None,
                    reason='Set LTD_TEST_AWS_ID, '
                           'LTD_TEST_AWS_SECRET and '
                           'LTD_TEST_BUCKET')
def test_delete_dir(request):
    session = boto3.session.Session(
        aws_access_key_id=os.getenv('LTD_TEST_AWS_ID'),
        aws_secret_access_key=os.getenv('LTD_TEST_AWS_SECRET'))
    s3 = session.resource('s3')
    bucket = s3.Bucket(os.getenv('LTD_TEST_BUCKET'))

    bucket_root = str(uuid.uuid4()) + '/'

    def cleanup():
        print("Cleaning up the bucket")
        delete_dir(os.getenv('LTD_TEST_BUCKET'),
                   bucket_root,
                   aws_access_key_id=os.getenv('LTD_TEST_AWS_ID'),
                   aws_secret_access_key=os.getenv('LTD_TEST_AWS_SECRET'))
    request.addfinalizer(cleanup)

    file_paths = ['a/test1.txt', 'a/b/test2.txt', 'a/b/c/test3.txt']
    upload_test_files(file_paths, bucket, bucket_root,
                      'sample-key', 'max-age=3600', 'text/plain')

    # Delete b/*
    delete_dir(os.getenv('LTD_TEST_BUCKET'),
               bucket_root + 'a/b/',
               aws_access_key_id=os.getenv('LTD_TEST_AWS_ID'),
               aws_secret_access_key=os.getenv('LTD_TEST_AWS_SECRET'))

    # Ensure paths outside of that are still available, but paths in b/ are
    # deleted
    bucket_paths = []
    for obj in bucket.objects.filter(Prefix=bucket_root):
        if obj.key.endswith('/'):
            continue
        bucket_paths.append(obj.key)

    for p in file_paths:
        bucket_path = bucket_root + p
        if p.startswith('a/b'):
            assert bucket_path not in bucket_paths
        else:
            assert bucket_path in bucket_paths

    # Attempt to delete an empty prefix. Ensure it does not raise an exception.
    delete_dir(os.getenv('LTD_TEST_BUCKET'),
               bucket_root + 'empty-prefix/',
               os.getenv('LTD_TEST_AWS_ID'),
               os.getenv('LTD_TEST_AWS_SECRET'))



"""S3 upload/sync utilities.
"""

__all__ = ('upload_dir', 'upload_file', 'upload_object',
           'create_dir_redirect_object', 'ObjectManager')

import os
import logging
import mimetypes

import boto3

from .exceptions import S3Error


def upload_dir(bucket_name, path_prefix, source_dir,
               upload_dir_redirect_objects=True,
               surrogate_key=None,
               surrogate_control=None, cache_control=None,
               acl=None,
               aws_access_key_id=None, aws_secret_access_key=None,
               aws_profile=None):
    """Upload a directory of files to S3.

    This function places the contents of the Sphinx HTML build directory
    into the ``/path_prefix/`` directory of an *existing* S3 bucket.
    Existing files on S3 are overwritten; files that no longer exist in the
    ``source_dir`` are deleted from S3.

    Parameters
    ----------
    bucket_name : `str`
        Name of the S3 bucket where documentation is uploaded.
    path_prefix : `str`
        The root directory in the bucket where documentation is stored.
    source_dir : `str`
        Path of the Sphinx HTML build directory on the local file system.
        The contents of this directory are uploaded into the ``/path_prefix/``
        directory of the S3 bucket.
    upload_dir_redirect_objects : `bool`, optional
        A feature flag to enable uploading objects to S3 for every directory.
        These objects contain ``x-amz-meta-dir-redirect=true`` HTTP headers
        that tell Fastly to issue a 301 redirect from the directory object to
        the `index.html`` in that directory.
    surrogate_key : `str`, optional
        The surrogate key to insert in the header of all objects
        in the ``x-amz-meta-surrogate-key`` field. This key is used to purge
        builds from the Fastly CDN when Editions change.
        If `None` then no header will be set.
    cache_control : `str`, optional
        This sets the ``Cache-Control`` header on the uploaded
        files. The ``Cache-Control`` header specifically dictates how content
        is cached by the browser (if ``surrogate_control`` is also set).
    surrogate_control : `str`, optional
        This sets the ``x-amz-meta-surrogate-control`` header
        on the uploaded files. The ``Surrogate-Control``
        or ``x-amz-meta-surrogate-control`` header is used in priority by
        Fastly to givern it's caching. This caching policy is *not* passed
        to the browser.
    acl : `str`, optional
        The pre-canned AWS access control list to apply to this upload.
        Can be ``'public-read'``, which allow files to be downloaded
        over HTTP by the public. See
        https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
        for an overview of S3's pre-canned ACL lists. Note that ACL settings
        are not validated locally. Default is `None`, meaning that no ACL
        is applied to an individual object. In this case, use ACLs applied
        to the bucket itself.
    aws_access_key_id : `str`, optional
        The access key for your AWS account. Also set
        ``aws_secret_access_key``.
    aws_secret_access_key : `str`, optional
        The secret key for your AWS account.
    aws_profile : `str`, optional
        Name of AWS profile in :file:`~/.aws/credentials`. Use this instead
        of ``aws_access_key_id`` and ``aws_secret_access_key`` for file-based
        credentials.

    Notes
    -----
    ``cache_control`` and  ``surrogate_control`` can be used together.
    ``surrogate_control`` takes priority in setting Fastly's POP caching,
    while ``cache_control`` then sets the browser's caching. For example:

    - ``cache_control='no-cache'``
    - ``surrogate_control='max-age=31536000'``

    together will ensure that the browser always does an ETAG server query,
    but that Fastly will cache the content for one year (or until purged).
    This configuration is good for files that are frequently changed in place.

    For immutable uploads simply using ``cache_control`` is more efficient
    since it allows the browser to also locally cache content.

    .. seelso:

       - `Fastly: Cache control tutorial
         <https://docs.fastly.com/guides/tutorials/cache-control-tutorial>`_.
       - `Google: HTTP caching <http://ls.st/39v>`_.
    """
    logger = logging.getLogger(__name__)

    logger.debug('s3upload.upload({0}, {1}, {2})'.format(
        bucket_name, path_prefix, source_dir))

    session = boto3.session.Session(
        profile_name=aws_profile,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key)
    s3 = session.resource('s3')
    bucket = s3.Bucket(bucket_name)

    metadata = {}
    if surrogate_key is not None:
        metadata['surrogate-key'] = surrogate_key
    if surrogate_control is not None:
        metadata['surrogate-control'] = surrogate_control

    manager = ObjectManager(session, bucket_name, path_prefix)

    for (rootdir, dirnames, filenames) in os.walk(source_dir):
        # name of root directory on S3 bucket
        bucket_root = os.path.relpath(rootdir, start=source_dir)
        if bucket_root in ('.', '/'):
            bucket_root = ''

        # Delete bucket directories that no longer exist in source
        bucket_dirnames = manager.list_dirnames_in_directory(bucket_root)
        for bucket_dirname in bucket_dirnames:
            if bucket_dirname not in dirnames:
                logger.debug(('Deleting bucket directory {0}'.format(
                    bucket_dirname)))
                manager.delete_directory(bucket_dirname)

        # Delete files that no longer exist in source
        bucket_filenames = manager.list_filenames_in_directory(bucket_root)
        for bucket_filename in bucket_filenames:
            if bucket_filename not in filenames:
                bucket_filename = os.path.join(bucket_root, bucket_filename)
                logger.debug(
                    'Deleting bucket file {0}'.format(bucket_filename))
                manager.delete_file(bucket_filename)

        # Upload files in directory
        for filename in filenames:
            local_path = os.path.join(rootdir, filename)
            bucket_path = os.path.join(path_prefix, bucket_root, filename)
            logger.debug('Uploading to {0}'.format(bucket_path))
            upload_file(local_path, bucket_path, bucket,
                        metadata=metadata, acl=acl,
                        cache_control=cache_control)

        # Upload a directory redirect object
        if upload_dir_redirect_objects is True:
            bucket_dir_path = os.path.join(path_prefix, bucket_root)
            create_dir_redirect_object(
                bucket_dir_path,
                bucket,
                metadata=metadata,
                acl=acl,
                cache_control=cache_control)


def upload_file(local_path, bucket_path, bucket,
                metadata=None, acl=None, cache_control=None):
    """Upload a file to the S3 bucket.

    This function uses the mimetypes module to guess and then set the
    Content-Type and Encoding-Type headers.

    Parameters
    ----------
    local_path : `str`
        Full path to a file on the local file system.
    bucket_path : `str`
        Destination path (also known as the key name) of the file in the
        S3 bucket.
    bucket : boto3 Bucket instance
        S3 bucket.
    metadata : `dict`, optional
        Header metadata values. These keys will appear in headers as
        ``x-amz-meta-*``.
    acl : `str`, optional
        A pre-canned access control list. See
        https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
        Default is `None`, mean that no ACL is applied to the object.
    cache_control : `str`, optional
        The cache-control header value. For example, ``'max-age=31536000'``.
    """
    logger = logging.getLogger(__name__)

    extra_args = {}
    if acl is not None:
        extra_args['ACL'] = acl
    if metadata is not None and len(metadata) > 0:  # avoid empty Metadata
        extra_args['Metadata'] = metadata
    if cache_control is not None:
        extra_args['CacheControl'] = cache_control

    # guess_type returns None if it cannot detect a type
    content_type, content_encoding = mimetypes.guess_type(local_path,
                                                          strict=False)
    if content_type is not None:
        extra_args['ContentType'] = content_type

    logger.debug(str(extra_args))

    obj = bucket.Object(bucket_path)
    # no return status from the upload_file api
    obj.upload_file(local_path, ExtraArgs=extra_args)


def upload_object(bucket_path, bucket, content='',
                  metadata=None, acl=None, cache_control=None,
                  content_type=None):
    """Upload an arbitrary object to an S3 bucket.

    Parameters
    ----------
    bucket_path : `str`
        Destination path (also known as the key name) of the file in the
        S3 bucket.
    content : `str` or `bytes`, optional
        Object content.
    bucket : boto3 Bucket instance
        S3 bucket.
    metadata : `dict`, optional
        Header metadata values. These keys will appear in headers as
        ``x-amz-meta-*``.
    acl : `str`, optional
        A pre-canned access control list. See
        https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
        Default is `None`, meaning that no ACL is applied to the object.
    cache_control : `str`, optional
        The cache-control header value. For example, ``'max-age=31536000'``.
    content_type : `str`, optional
        The object's content type (such as ``text/html``). If left unset,
        no MIME type is passed to boto3 (which defaults to
        ``binary/octet-stream``).
    """
    obj = bucket.Object(bucket_path)

    # Object.put seems to be sensitive to None-type kwargs, so we filter first
    args = {}
    if metadata is not None and len(metadata) > 0:  # avoid empty Metadata
        args['Metadata'] = metadata
    if acl is not None:
        args['ACL'] = acl
    if cache_control is not None:
        args['CacheControl'] = cache_control
    if content_type is not None:
        args['ContentType'] = content_type

    obj.put(Body=content, **args)


def create_dir_redirect_object(bucket_dir_path, bucket, metadata=None,
                               acl=None, cache_control=None):
    """Create an S3 object representing a directory that's designed to
    redirect a browser (via Fastly) to the ``index.html`` contained inside
    that directory.

    Parameters
    ----------
    bucket_dir_path : `str`
        Full name of the object in the S3, which is equivalent to a POSIX
        directory path, like ``dir1/dir2``.
    bucket : boto3 Bucket instance
        S3 bucket.
    metadata : `dict`, optional
        Header metadata values. These keys will appear in headers as
        ``x-amz-meta-*``.
    acl : `str`, optional
        A pre-canned access control list. See
        https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
        Default is None, meaning that no ACL is applied to the object.
    cache_control : `str`, optional
        The cache-control header value. For example, ``'max-age=31536000'``.
    """
    # Just the name of the 'directory' itself
    bucket_dir_path = bucket_dir_path.rstrip('/')

    # create a copy of metadata
    if metadata is not None:
        metadata = dict(metadata)
    else:
        metadata = {}

    # header used by LTD's Fastly Varnish config to create a 301 redirect
    metadata['dir-redirect'] = 'true'

    upload_object(bucket_dir_path,
                  content='',
                  bucket=bucket,
                  metadata=metadata,
                  acl=acl,
                  cache_control=cache_control)


class ObjectManager(object):
    """Manage objects existing in a bucket under a specific ``bucket_root``.

    The ObjectManager maintains information about objects that exist in the
    bucket, and can delete objects that no longer exist in the source.

    Parameters
    ----------
    session : :class:`boto3.session.Session`
        A boto3 session instance provisioned with the correct identities.
    bucket_name : `str`
        Name of the S3 bucket.
    bucket_root : `str`
        The version slug is the name root directory in the bucket where
        documentation is stored.
    """
    def __init__(self, session, bucket_name, bucket_root):
        super().__init__()
        self._logger = logging.getLogger(__name__)

        s3 = session.resource('s3')
        bucket = s3.Bucket(bucket_name)
        self._session = session
        self._bucket = bucket
        self._bucket_root = bucket_root
        # Strip trailing '/' from bucket_root for comparisons
        self._bucket_root = self._bucket_root.rstrip('/')

    def list_filenames_in_directory(self, dirname):
        """List all file-type object names that exist at the root of this
        bucket directory.

        Parameters
        ----------
        dirname : `str`
            Directory name in the bucket relative to ``bucket_root/``.

        Returns
        -------
        filenames : `list`
            List of file names (`str`), relative to ``bucket_root/``, that
            exist at the root of ``dirname``.
        """
        prefix = self._create_prefix(dirname)
        filenames = []
        for obj in self._bucket.objects.filter(Prefix=prefix):
            if obj.key.endswith('/'):
                # a directory redirect object, not a file
                continue
            obj_dirname = os.path.dirname(obj.key)
            if obj_dirname == prefix:
                # object is at root of directory
                filenames.append(os.path.relpath(obj.key,
                                                 start=prefix))
        return filenames

    def list_dirnames_in_directory(self, dirname):
        """List all names of directories that exist at the root of this
        bucket directory.

        Note that *directories* don't exist in S3; rather directories are
        inferred from path names.

        Parameters
        ----------
        dirname : `str`
            Directory name in the bucket relative to ``bucket_root``.

        Returns
        -------
        dirnames : `list`
            List of directory names (`str`), relative to ``bucket_root/``,
            that exist at the root of ``dirname``.
        """
        prefix = self._create_prefix(dirname)
        dirnames = []
        for obj in self._bucket.objects.filter(Prefix=prefix):
            # get directory name of every object under this path prefix
            dirname = os.path.dirname(obj.key)

            # dirname is empty if the object happens to be the directory
            # redirect object object for the prefix directory (directory
            # redirect objects are named after directories and have metadata
            # that tells Fastly to redirect the browser to the index.html
            # contained in the directory).
            if dirname == '':
                dirname = obj.key + '/'

            # Strip out the path prefix from the directory name
            rel_dirname = os.path.relpath(dirname, start=prefix)

            # If there's only one part then this directory is at the root
            # relative to the prefix. We want this.
            dir_parts = rel_dirname.split('/')
            if len(dir_parts) == 1:
                dirnames.append(dir_parts[0])

        # Above algorithm finds root directories for all *files* in sub
        # subdirectories; trim down to the unique set.
        dirnames = list(set(dirnames))

        # Remove posix-like relative directory names that can appear
        # in the bucket listing.
        for filtered_dir in ('.', '..'):
            if filtered_dir in dirnames:
                dirnames.remove(filtered_dir)

        return dirnames

    def _create_prefix(self, dirname):
        """Make an absolute directory path in the bucker for dirname,
        which is is assumed relative to the self._bucket_root prefix directory.
        """
        if dirname in ('.', '/'):
            dirname = ''
        # Strips trailing slash from dir prefix for comparisons
        # os.path.dirname() returns directory names without a trailing /
        prefix = os.path.join(self._bucket_root, dirname)
        prefix = prefix.rstrip('/')
        return prefix

    def delete_file(self, filename):
        """Delete a file from the bucket.

        Parameters
        ----------
        filename : `str`
            Name of the file, relative to ``bucket_root/``.
        """
        key = os.path.join(self._bucket_root, filename)
        objects = list(self._bucket.objects.filter(Prefix=key))
        for obj in objects:
            obj.delete()

    def delete_directory(self, dirname):
        """Delete a directory (and contents) from the bucket.

        Parameters
        ----------
        dirname : `str`
            Name of the directory, relative to ``bucket_root/``.

        Raises
        ------
        RuntimeError
            Raised when there are no objects to delete (directory
            does not exist).
        """
        key = os.path.join(self._bucket_root, dirname)
        if not key.endswith('/'):
            key += '/'

        key_objects = [{'Key': obj.key}
                       for obj in self._bucket.objects.filter(Prefix=key)]
        if len(key_objects) == 0:
            msg = 'No objects in bucket directory {}'.format(dirname)
            raise RuntimeError(msg)
        delete_keys = {'Objects': key_objects}
        # based on http://stackoverflow.com/a/34888103
        s3 = self._session.resource('s3')
        r = s3.meta.client.delete_objects(Bucket=self._bucket.name,
                                          Delete=delete_keys)
        self._logger.debug(r)
        if 'Errors' in r:
            raise S3Error('S3 could not delete {0}'.format(key))



"""Test utilities available to test modules.
"""

__all__ = ('upload_test_files',)

import os
from tempfile import TemporaryDirectory


def upload_test_files(file_paths, bucket, bucket_root,
                      surrogate_key, cache_control, content_type):
    r"""Create and upload files to S3 as specified by their paths alone.

    This is useful for specifying a filesystem tree, and seeing if Conveyor's
    S3 interactions maintain or manipulate that tree properly.

    Parameters
    ----------
    file_paths : `list`
        List of file paths that specify a directory tree. For example:

        .. code-block:: py

           ['a.txt', 'b/c.txt', 'b/d/e.txt']
    bucket : `S3.Bucket`
        A boto3 bucket built from a ``boto3.session.Session``\ 's ``resource``
        method. Test files will be uploaded to this bucket.
    bucket_root : `str`
        Directory prefix in the bucket for testing.
    surrogate_key : `str`
        A Fastly CDN surrogate key for testing.
    cache_control : `str`
        Cache-control setting, such as ``"max-age=3600"``.
    content_type : `str`
        Content-type header of the test files, such as ``"text/plain"``.
    """
    with TemporaryDirectory() as temp_dir:
        for p in file_paths:
            full_path = os.path.join(temp_dir, p)
            full_dir = os.path.dirname(full_path)
            os.makedirs(full_dir, exist_ok=True)
            with open(full_path, 'w') as f:
                f.write('content')

            extra_args = {
                'Metadata': {'surrogate-key': surrogate_key},
                'ContentType': content_type,
                'CacheControl': cache_control}
            obj = bucket.Object(bucket_root + p)
            obj.upload_file(full_path, ExtraArgs=extra_args)



"""S3 utilities.
"""

__all__ = ('open_bucket',)

import boto3


def open_bucket(bucket_name,
                aws_access_key_id=None, aws_secret_access_key=None,
                aws_profile=None):
    """Open an S3 Bucket resource.

    Parameters
    ----------
    bucket_name : `str`
        Name of the S3 bucket.
    aws_access_key_id : `str`, optional
        The access key for your AWS account. Also set
        ``aws_secret_access_key``.
    aws_secret_access_key : `str`, optional
        The secret key for your AWS account.
    aws_profile : `str`, optional
        Name of AWS profile in :file:`~/.aws/credentials`. Use this instead
        of ``aws_access_key_id`` and ``aws_secret_access_key`` for file-based
        credentials.

    Returns
    -------
    bucket : Boto3 S3 Bucket instance
        The S3 bucket as a Boto3 instance.
    """
    session = boto3.session.Session(
        profile_name=aws_profile,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key)
    s3 = session.resource('s3')
    bucket = s3.Bucket(bucket_name)
    return bucket



"""Copy an S3 directory to another prefix in the same bucket.
"""

__all__ = ('copy_dir',)

import os
import boto3

from .delete import delete_dir


def copy_dir(bucket_name, src_path, dest_path,
             aws_access_key_id=None, aws_secret_access_key=None,
             aws_profile=None,
             surrogate_key=None, cache_control=None,
             surrogate_control=None,
             create_directory_redirect_object=True):
    """Copy objects from one directory in a bucket to another directory in
    the same bucket.

    Object metadata is preserved while copying, with the following exceptions:

    - If a new surrogate key is provided it will replace the original one.
    - If ``cache_control`` and ``surrogate_control`` values are provided they
      will replace the old one.

    Parameters
    ----------
    bucket_name : `str`
        Name of an S3 bucket.
    src_path : `str`
        Source directory in the S3 bucket. The ``src_path`` should ideally end
        in a trailing `'/'`. E.g. `'dir/dir2/'`.
    dest_path : `str`
        Destination directory in the S3 bucket. The ``dest_path`` should
        ideally end in a trailing `'/'`. E.g. `'dir/dir2/'`. The destination
        path cannot contain the source path.
    aws_access_key_id : `str`
        The access key for your AWS account. Also set
        ``aws_secret_access_key``.
    aws_secret_access_key : `str`
        The secret key for your AWS account.
    aws_profile : `str`, optional
        Name of AWS profile in :file:`~/.aws/credentials`. Use this instead
        of ``aws_access_key_id`` and ``aws_secret_access_key`` for file-based
        credentials.
    surrogate_key : `str`, optional
        The surrogate key to insert in the header of all objects in the
        ``x-amz-meta-surrogate-key`` field. This key is used to purge
        builds from the Fastly CDN when Editions change.
        If `None` then no header will be set.
        If the object already has a ``x-amz-meta-surrogate-key`` header then
        it will be replaced.
    cache_control : `str`, optional
        This sets (and overrides) the ``Cache-Control`` header on the copied
        files. The ``Cache-Control`` header specifically dictates how content
        is cached by the browser (if ``surrogate_control`` is also set).
    surrogate_control : `str`, optional
        This sets (and overrides) the ``x-amz-meta-surrogate-control`` header
        on the copied files. The ``Surrogate-Control``
        or ``x-amz-meta-surrogate-control`` header is used in priority by
        Fastly to givern it's caching. This caching policy is *not* passed
        to the browser.
    create_directory_redirect_object : `bool`, optional
        Create a directory redirect object for the root directory. The
        directory redirect object is an empty S3 object named after the
        directory (without a trailing slash) that contains a
        ``x-amz-meta-dir-redirect=true`` HTTP header. LSST the Docs' Fastly
        VCL is configured to redirect requests for a directory path to the
        directory's ``index.html`` (known as *courtesy redirects*).

    Raises
    ------
    ltdconveyor.s3.S3Error
        Thrown by any unexpected faults from the S3 API.
    RuntimeError
        Thrown when the source and destination directories are the same.
    """
    if not src_path.endswith('/'):
        src_path += '/'
    if not dest_path.endswith('/'):
        dest_path += '/'

    # Ensure the src_path and dest_path don't contain each other
    common_prefix = os.path.commonprefix([src_path, dest_path])
    if common_prefix == src_path:
        msg = 'Common prefix {0} is same as source dir {1}'.format(
            common_prefix, src_path)
        raise RuntimeError(msg)
    if common_prefix == dest_path:
        msg = 'Common prefix {0} is same as dest dir {1}'.format(
            common_prefix, dest_path)
        raise RuntimeError(msg)

    # Delete any existing objects in the destination
    delete_dir(bucket_name, dest_path,
               aws_access_key_id, aws_secret_access_key)

    session = boto3.session.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        profile_name=aws_profile)
    s3 = session.resource('s3')
    bucket = s3.Bucket(bucket_name)

    # Copy each object from source to destination
    for src_obj in bucket.objects.filter(Prefix=src_path):
        src_rel_path = os.path.relpath(src_obj.key, start=src_path)
        dest_key_path = os.path.join(dest_path, src_rel_path)

        # the src_obj (ObjectSummary) doesn't include headers afaik
        head = s3.meta.client.head_object(Bucket=bucket_name,
                                          Key=src_obj.key)
        metadata = head['Metadata']
        content_type = head['ContentType']

        # try to use original Cache-Control header if new one is not set
        if cache_control is None and 'CacheControl' in head:
            cache_control = head['CacheControl']

        if surrogate_control is not None:
            metadata['surrogate-control'] = surrogate_control

        if surrogate_key is not None:
            metadata['surrogate-key'] = surrogate_key

        s3.meta.client.copy_object(
            Bucket=bucket_name,
            Key=dest_key_path,
            CopySource={'Bucket': bucket_name, 'Key': src_obj.key},
            MetadataDirective='REPLACE',
            Metadata=metadata,
            ACL='public-read',
            CacheControl=cache_control,
            ContentType=content_type)

    if create_directory_redirect_object:
        dest_dirname = dest_path.rstrip('/')
        obj = bucket.Object(dest_dirname)
        metadata = {'dir-redirect': 'true'}
        obj.put(Body='',
                ACL='public-read',
                Metadata=metadata,
                CacheControl=cache_control)



"""Command-line interface ``ltd`` as a Click application.
"""

__all__ = ('main',)

import logging

import click

from .upload import upload


# Add -h as a help shortcut option
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    '--log-level', 'log_level',
    type=click.Choice(['warning', 'info', 'debug']),
    default='info',
    help='Logging level (for first-party messages). Default: `info`.'
)
@click.option(
    '--host', 'keeper_hostname',
    default='https://keeper.lsst.codes',
    envvar='LTD_HOST',
    help='Hostname of the LTD Keeper API (or `$LTD_HOST`). '
         'Default: `https://keeper.lsst.codes`.'
)
@click.option(
    '-u', '--user', 'username',
    envvar='LTD_USERNAME',
    help='Username for LTD Keeper (or `$LTD_USERNAME`).'
)
@click.option(
    '-p', '--password', 'password',
    envvar='LTD_PASSWORD',
    help='Password for LTD Keeper (or `$LTD_PASSWORD`).'
)
@click.version_option()
@click.pass_context
def main(ctx, log_level, keeper_hostname, username, password):
    """ltd is a command-line client for LSST the Docs.

    Use ltd to upload new site builds, and to work with the LTD Keeper API.
    """
    ch = logging.StreamHandler()
    if log_level in ('warning', 'info'):
        formatter = logging.Formatter(
            '%(asctime)s %(levelname)8s | %(message)s')
    else:
        formatter = logging.Formatter(
            '%(asctime)s %(levelname)8s %(name)s | %(message)s')
    ch.setFormatter(formatter)

    logger = logging.getLogger('ltdconveyor')
    logger.addHandler(ch)
    logger.setLevel(log_level.upper())

    # Subcommands should use the click.pass_obj decorator to get this
    # ctx.obj object as the first argument.
    ctx.obj = {
        'keeper_hostname': keeper_hostname,
        'username': username,
        'password': password,
        'token': None
    }


@main.command()
@click.argument('topic', default=None, required=False, nargs=1)
@click.pass_context
def help(ctx, topic, **kw):
    """Show help for any command.
    """
    # The help command implementation is taken from
    # https://www.burgundywall.com/post/having-click-help-subcommand
    if topic is None:
        click.echo(ctx.parent.get_help())
    else:
        click.echo(main.commands[topic].get_help(ctx))


# Add subcommands from other modules
main.add_command(upload)



"""ltd upload subcommand.
"""

__all__ = ('upload',)

import logging
import os
import re
import sys
from pathlib import Path

import click

from ..keeper.build import register_build, confirm_build
from ltdconveyor.s3.presignedpost import (prescan_directory, upload_dir,
                                          upload_directory_objects)
from .utils import ensure_login


@click.command()
@click.option(
    '--product',
    required=True,
    help='Product name.'
)
@click.option(
    '--dir', 'dirname',
    default='.',
    type=click.Path(exists=False, file_okay=False, dir_okay=True),
    help='Directory with files to upload. Default: `.` (current working '
         'directory).'
)
@click.option(
    '--git-ref',
    help='Git ref, or space-delimited list of git refs. This versions the '
         'build and helps LTD Keeper assign the build to an edition. '
         'Alternatively, version information can be auto-discovered by '
         'setting --travis in a Travis CI job or --gh for GitHub actions.'
)
@click.option(
    '--gh', 'ci_env', flag_value='gh',
    help='Use environment variables from a GitHub Actions environment to set '
         'the --git-ref option.'
)
@click.option(
    '--travis', 'ci_env', flag_value='travis',
    help='Use environment variables from a Travis CI environment to set '
         'the --git-ref option.'
)
@click.option(
    '--on-travis-push/--no-travis-push', 'on_travis_push', default=True,
    help='Upload on a Travis CI push (enabled by default). Must be used with '
         '--travis.'
)
@click.option(
    '--on-travis-pr/--no-travis-pr', 'on_travis_pr', default=False,
    help='Upload on a Travis CI pull request event (disabled by default). '
         'Must be used with --travis.'
)
@click.option(
    '--on-travis-api/--no-travis-api', 'on_travis_api', default=True,
    help='Upload on a Travis CI API event (enabled by default). '
         'Must be used with --travis.'
)
@click.option(
    '--on-travis-cron/--no-travis-cron', 'on_travis_cron', default=True,
    help='Upload on a Travis CI cron event (enabled by default). '
         'Must be used with --travis.'
)
@click.option(
    '--skip', 'skip_upload', default=False, is_flag=True,
    envvar='LTD_SKIP_UPLOAD',
    help='Skip the upload, making the command a no-op. '
         'Useful in CI environments to disable a site upload just by setting '
         'this option or the environment variable $LTD_SKIP_UPLOAD=true.'
)
@click.pass_context
def upload(ctx, product, git_ref, dirname, ci_env, on_travis_push,
           on_travis_pr, on_travis_api, on_travis_cron, skip_upload):
    """Upload a new site build to LSST the Docs.
    """
    logger = logging.getLogger(__name__)

    if skip_upload:
        click.echo('Skipping ltd upload.')
        sys.exit(0)

    logger.debug('CI environment: %s', ci_env)
    logger.debug('Travis events settings. '
                 'On Push: %r, PR: %r, API: %r, Cron: %r',
                 on_travis_push, on_travis_pr, on_travis_api, on_travis_cron)

    # Abort upload on Travis CI under certain events
    if ci_env == 'travis' and \
            _should_skip_travis_event(
                on_travis_push, on_travis_pr, on_travis_api, on_travis_cron):
        sys.exit(0)

    # Authenticate to LTD Keeper host
    ensure_login(ctx)

    # Detect git refs
    git_refs = _get_git_refs(ci_env, git_ref)

    # Prescan directory names. Each directory needs a presigned POST URL.
    base_dir = Path(dirname)
    dirnames = prescan_directory(base_dir)

    build_resource = register_build(
        ctx.obj['keeper_hostname'],
        ctx.obj['token'],
        product,
        git_refs,
        dirnames=dirnames
    )
    logger.debug('Created build resource %r', build_resource)

    # Do the upload.
    upload_dir(
        post_urls=build_resource['post_prefix_urls'],
        base_dir=base_dir
    )
    logger.debug('Upload complete for %r', build_resource['self_url'])

    # Upload directory objects for redirects
    upload_directory_objects(
        post_urls=build_resource['post_dir_urls'],
    )

    # Confirm upload
    confirm_build(
        build_resource['self_url'],
        ctx.obj['token']
    )
    logger.info('Build %r complete', build_resource['self_url'])
    logger.info('Published build URL: %s', build_resource['published_url'])


def _should_skip_travis_event(on_travis_push, on_travis_pr, on_travis_api,
                              on_travis_cron):
    """Detect if the upload should be skipped based on the
    ``TRAVIS_EVENT_TYPE`` environment variable.

    Returns
    -------
    should_skip : `bool`
        True if the upload should be skipped based on the combination of
        ``TRAVIS_EVENT_TYPE`` and user settings.
    """
    travis_event = os.getenv('TRAVIS_EVENT_TYPE')
    if travis_event is None:
        raise click.UsageError(
            'Using --travis but the TRAVIS_EVENT_TYPE '
            'environment variable is not detected.')

    if travis_event == 'push' and on_travis_push is False:
        click.echo('Skipping upload on Travis push event.')
        return True
    elif travis_event == 'pull_request' and on_travis_pr is False:
        click.echo('Skipping upload on Travis pull request event.')
        return True
    elif travis_event == 'api' and on_travis_api is False:
        click.echo('Skipping upload on Travis pull request event.')
        return True
    elif travis_event == 'cron' and on_travis_cron is False:
        click.echo('Skipping upload on Travis cron event.')
        return True
    else:
        return False


def _get_git_refs(ci_env, user_git_ref):
    if ci_env == 'travis' and user_git_ref is None:
        # Get git refs from Travis environment
        git_refs = _get_travis_git_refs()
    elif ci_env == 'gh' and user_git_ref is None:
        # Get git refs from GitHub Actions environment
        git_refs = _get_gh_actions_git_refs()
    elif user_git_ref is not None:
        # Get git refs from command line
        git_refs = user_git_ref.split()
    else:
        raise click.UsageError('--git-ref is required.')
    return git_refs


def _get_travis_git_refs():
    git_refs = [os.getenv('TRAVIS_BRANCH')]
    if git_refs[0] is None:
        raise click.UsageError(
            'Using --travis but the TRAVIS_BRANCH environment variable is '
            'not detected.')
    return git_refs


def _get_gh_actions_git_refs():
    github_ref = os.getenv('GITHUB_REF')
    if github_ref is None:
        raise click.UsageError(
            'Using --gh but the GITHUB_REF environment variable is '
            'not detected.')
    match = re.match(
        r"refs/(heads|tags|pull)/(?P<ref>.+)",
        github_ref
    )
    if not match:
        raise click.UsageError(
            'Could not parse the GITHUB_REF environment variable: {0}'.format(
                github_ref)
        )
    ref = match.group("ref")
    return [ref]



"""Tests for `ltdconveyor.fastly`.
"""

import uuid
import pytest
import responses

from ltdconveyor.fastly import purge_key, FastlyError


@responses.activate
def test_purge_key():
    service_id = 'SU1Z0isxPaozGVKXdv0eY'
    api_key = 'd3cafb4dde4dbeef'
    surrogate_key = uuid.uuid4().hex

    url = 'https://api.fastly.com/service/{0}/purge/{1}'.format(
        service_id, surrogate_key)

    # Mock the API call and response
    responses.add(responses.POST, url, status=200)

    purge_key(surrogate_key, service_id, api_key)
    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == url
    assert responses.calls[0].request.headers['Fastly-Key'] == api_key
    assert responses.calls[0].request.headers['Accept'] == 'application/json'


@responses.activate
def test_purge_key_fail():
    service_id = 'SU1Z0isxPaozGVKXdv0eY'
    api_key = 'd3cafb4dde4dbeef'
    surrogate_key = uuid.uuid4().hex

    url = 'https://api.fastly.com/service/{0}/purge/{1}'.format(
        service_id, surrogate_key)

    # Mock the API call and response
    responses.add(responses.POST, url, status=404)

    with pytest.raises(FastlyError):
        purge_key(surrogate_key, service_id, api_key)
    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == url
    assert responses.calls[0].request.headers['Fastly-Key'] == api_key
    assert responses.calls[0].request.headers['Accept'] == 'application/json'



"""Management of Fastly CDN caching.

See https://docs.fastly.com/api for background.
"""

__all__ = ('purge_key', 'FastlyError')

import logging
import requests

from .exceptions import ConveyorError


def purge_key(surrogate_key, service_id, api_key):
    """Instant purge URLs with a given surrogate key from the Fastly caches.

    Parameters
    ----------
    surrogate_key : `str`
        Surrogate key header (``x-amz-meta-surrogate-key``) value of objects
        to purge from the Fastly cache.
    service_id : `str`
        Fastly service ID.
    api_key : `str`
        Fastly API key.

    Raises
    ------
    FastlyError
       Error with the Fastly API usage.

    Notes
    -----
    This function uses Fastly's ``/service/{service}/purge/{key}`` endpoint.
    See the `Fastly Purge documentation <http://ls.st/jxg>`_ for more
    information.

    For other Fastly APIs, consider using `fastly-py
    <https://github.com/fastly/fastly-py>`_.
    """
    logger = logging.getLogger(__name__)

    api_root = 'https://api.fastly.com'
    path = '/service/{service}/purge/{surrogate_key}'.format(
        service=service_id,
        surrogate_key=surrogate_key)
    logger.info('Fastly purge {0}'.format(path))
    r = requests.post(api_root + path,
                      headers={'Fastly-Key': api_key,
                               'Accept': 'application/json'})
    if r.status_code != 200:
        raise FastlyError(r.json)


class FastlyError(ConveyorError):
    """Error related to Fastly API usage.
    """



"""Helpers for the CLI (not click commands).
"""

import logging
import sys

import click

from ..keeper.login import get_keeper_token


def ensure_login(ctx):
    """Ensure a token is in the Click context object or authenticate and obtain
    the token from LTD Keeper.

    Parameters
    ----------
    ctx : `click.Context`
        The Click context. ``ctx.obj`` must be a `dict` that contains keys:
        ``keeper_hostname``, ``username``, ``password``, ``token``. This
        context object is prepared by the main Click group,
        `ltdconveyor.cli.main.main`.
    """
    logger = logging.getLogger(__name__)

    if ctx.obj['token'] is None:
        if ctx.obj['username'] is None or ctx.obj['password'] is None:
            raise click.UsageError(
                'Use `ltd -u <username> -p <password> COMMAND` to '
                'authenticate to the LTD Keeper server.')
            sys.exit(1)

        logger.debug(
            'About to get token for user %s at %s',
            ctx.obj['username'],
            ctx.obj['keeper_hostname'])

        token = get_keeper_token(
            ctx.obj['keeper_hostname'],
            ctx.obj['username'],
            ctx.obj['password'])
        ctx.obj['token'] = token

        logger.debug(
            'Got token for user %s at %s',
            ctx.obj['username'],
            ctx.obj['keeper_hostname'])

    else:
        logger.debug(
            'Token already exists.')



"""Tests for the ``ltdconveyor.s3.presignedpost`` module.
"""

import sys
import cgi
from io import BytesIO
from pathlib import Path, PurePosixPath
from unittest.mock import call

import pytest
import responses

from ltdconveyor.exceptions import ConveyorError
from ltdconveyor.s3.presignedpost import (
    format_relative_dirname, prescan_directory, upload_dir, upload_file,
    upload_directory_objects)
from ltdconveyor.s3.exceptions import S3Error


@pytest.mark.parametrize(
    'directory,basedir,expected',
    [
        ('/home/project/sub/dir', '/home/project/', 'sub/dir/'),
        ('/home/project/sub/dir', '/home/project', 'sub/dir/'),
        ('/home/project/sub/dir/', '/home/project', 'sub/dir/'),
        ('/home/project/', '/home/project/', '/'),
    ]
)
def test_relative_dirname(directory, basedir, expected):
    directory = PurePosixPath(directory)
    basedir = PurePosixPath(basedir)
    result = format_relative_dirname(directory, basedir)
    assert result == expected


def test_prescan_directory():
    """Test the prescan_directory function using a test directory in
    ``tests/data/test-site/``
    """
    base_dir = Path(__file__).parent / 'data/test-site'
    assert base_dir.is_dir()

    dirs = prescan_directory(base_dir)
    assert '/' in dirs
    assert 'a/' in dirs
    assert 'a/aa/' in dirs
    assert 'b/' in dirs
    assert len(dirs) == 4


@pytest.mark.xfail(
    reason='assert_has_calls fails for unknown reason on Travis')
def test_upload_dir(mocker):
    mock_upload_file = mocker.patch('ltdconveyor.s3.presignedpost.upload_file')

    post_urls = {
        '/': {
            'url': 'https://example.com',
            'fields': {'key': '/prefix/{$filename}'}
        },
        'a/': {
            'url': 'https://example.com',
            'fields': {'key': '/prefix/a/{$filename}'}
        },
        'a/aa/': {
            'url': 'https://example.com',
            'fields': {'key': '/prefix/a/aa/{$filename}'}
        },
        'b/': {
            'url': 'https://example.com',
            'fields': {'key': '/prefix/b/{$filename}'}
        }
    }
    base_dir = Path(__file__).parent / 'data/test-site'

    upload_dir(post_urls=post_urls, base_dir=base_dir)

    assert mock_upload_file.call_count == 4
    upload_file_calls = [
        call(local_path=base_dir / 'index.html',
             post_url=post_urls['/']['url'],
             post_fields=post_urls['/']['fields']),
        call(local_path=base_dir / 'a/index.html',
             post_url=post_urls['a/']['url'],
             post_fields=post_urls['a/']['fields']),
        call(local_path=base_dir / 'a/aa/index.html',
             post_url=post_urls['a/aa/']['url'],
             post_fields=post_urls['a/aa/']['fields']),
        call(local_path=base_dir / 'b/index.html',
             post_url=post_urls['b/']['url'],
             post_fields=post_urls['b/']['fields']),
    ]
    mock_upload_file.assert_has_calls(upload_file_calls)


def test_upload_dir_bad_posturls(mocker):
    """Test upload_dir when post_urls do not include the necessary directory.
    """
    mocker.patch('ltdconveyor.s3.presignedpost.upload_file')

    post_urls = {}
    base_dir = Path(__file__).parent / 'data/test-site'

    with pytest.raises(ConveyorError):
        upload_dir(post_urls=post_urls, base_dir=base_dir)


@responses.activate
def test_upload_file():
    local_path = Path(__file__).parent / 'data/test-site/index.html'
    post_url = 'https://example.com'
    post_fields = {'key': 'bucket/base/${filename}'}

    responses.add(responses.POST, 'https://example.com', status=204)

    upload_file(
        local_path=local_path,
        post_url=post_url,
        post_fields=post_fields)

    call = responses.calls[0]
    mimetype, options = cgi.parse_header(call.request.headers['Content-Type'])
    assert mimetype == 'multipart/form-data'
    pdict = {'boundary': options['boundary'].encode('ascii'),
             'CONTENT-LENGTH': call.request.headers['Content-Length']}
    data = BytesIO(responses.calls[0].request.body)
    parsed_body = cgi.parse_multipart(data, pdict)

    if sys.version_info[:3] >= (3, 7, 0):
        assert parsed_body['Content-Type'] == ['text/html']
        assert parsed_body['key'] == ['bucket/base/${filename}']
    else:
        assert parsed_body['Content-Type'] == [b'text/html']
        assert parsed_body['key'] == [b'bucket/base/${filename}']


@responses.activate
def test_upload_file_failed():
    local_path = Path(__file__).parent / 'data/test-site/index.html'
    post_url = 'https://example.com'
    post_fields = {'key': 'bucket/base/${filename}'}

    responses.add(responses.POST, 'https://example.com', status=403)

    with pytest.raises(S3Error):
        upload_file(
            local_path=local_path,
            post_url=post_url,
            post_fields=post_fields)


@responses.activate
def test_upload_directory_objects():
    post_urls = {
        '/': {
            'url': 'https://example.com',
            'fields': {'key': 'bucket/base'}
        },
        'a/': {
            'url': 'https://example.com',
            'fields': {'key': 'bucket/base/a'}
        },
        'a/aa/': {
            'url': 'https://example.com',
            'fields': {'key': 'bucket/base/a/aa'}
        },
        'b/': {
            'url': 'https://example.com',
            'fields': {'key': 'bucket/base/b/{$filename}'}
        }
    }
    responses.add(responses.POST, 'https://example.com', status=204)

    upload_directory_objects(post_urls=post_urls)

    assert len(responses.calls) == 4



"""Tests for s3copy.

These tests required the following environment variables:

``LTD_TEST_AWS_ID``
    AWS access key ID

``LTD_TEST_AWS_SECRET``
    AWS secret access key

``LTD_TEST_BUCKET``
    Name of an S3 bucket that already exists and can be used for testing.

The tests will be skipped if they are not available.

Note that this test will create a random uuid4) directory at the root of
``LTD_TEST_BUCKET``, though the test harness will attempt to delete
it.
"""

import os
import uuid

import boto3
import pytest

from ltdconveyor.s3 import copy_dir, delete_dir
from ltdconveyor.testutils import upload_test_files


@pytest.mark.skipif(os.getenv('LTD_TEST_AWS_ID') is None or
                    os.getenv('LTD_TEST_AWS_SECRET') is None or
                    os.getenv('LTD_TEST_BUCKET') is None,
                    reason='Set LTD_TEST_AWS_ID, '
                           'LTD_TEST_AWS_SECRET and '
                           'LTD_TEST_BUCKET')
def test_copy_directory(request):
    session = boto3.session.Session(
        aws_access_key_id=os.getenv('LTD_TEST_AWS_ID'),
        aws_secret_access_key=os.getenv('LTD_TEST_AWS_SECRET'))
    s3 = session.resource('s3')
    bucket = s3.Bucket(os.getenv('LTD_TEST_BUCKET'))

    bucket_root = str(uuid.uuid4()) + '/'

    def cleanup():
        print("Cleaning up the bucket")
        delete_dir(os.getenv('LTD_TEST_BUCKET'),
                   bucket_root,
                   aws_access_key_id=os.getenv('LTD_TEST_AWS_ID'),
                   aws_secret_access_key=os.getenv('LTD_TEST_AWS_SECRET'))
    request.addfinalizer(cleanup)

    initial_paths = ['test1.txt', 'test2.txt', 'aa/test3.txt']
    new_paths = ['test4.txt', 'bb/test5.txt']

    # add old and new file sets
    upload_test_files(initial_paths, bucket, bucket_root + 'a/',
                      'sample-key', 'max-age=3600', 'text/plain')
    upload_test_files(new_paths, bucket, bucket_root + 'b/',
                      'sample-key', 'max-age=3600', 'text/plain')

    # copy files
    copy_dir(
        bucket_name=os.getenv('LTD_TEST_BUCKET'),
        src_path=bucket_root + 'b/',
        dest_path=bucket_root + 'a/',
        aws_access_key_id=os.getenv('LTD_TEST_AWS_ID'),
        aws_secret_access_key=os.getenv('LTD_TEST_AWS_SECRET'),
        surrogate_key='new-key',
        surrogate_control='max-age=31536000',
        cache_control='no-cache')

    # Test files in the a/ directory are from b/
    for obj in bucket.objects.filter(Prefix=bucket_root + 'a/'):
        bucket_path = os.path.relpath(obj.key, start=bucket_root + 'a/')
        assert bucket_path in new_paths
        # ensure correct metadata
        head = s3.meta.client.head_object(
            Bucket=os.getenv('LTD_TEST_BUCKET'),
            Key=obj.key)
        assert head['CacheControl'] == 'no-cache'
        assert head['ContentType'] == 'text/plain'
        assert head['Metadata']['surrogate-key'] == 'new-key'
        assert head['Metadata']['surrogate-control'] == 'max-age=31536000'

    # Test that a directory object exists
    bucket_paths = [obj.key
                    for obj in bucket.objects.filter(Prefix=bucket_root + 'a')]
    assert os.path.join(bucket_root, 'a') in bucket_paths


def test_copy_dir_src_in_dest():
    """Test that copy_directory fails raises a RuntimeError if source in
    destination.
    """
    with pytest.raises(RuntimeError):
        copy_dir('example', 'dest/src', 'dest',
                 aws_access_key_id='id',
                 aws_secret_access_key='key')


def test_copy_dir_dest_in_src():
    """Test that copy_directory fails raises a RuntimeError if destination
    is part of the source.
    """
    with pytest.raises(RuntimeError):
        copy_dir('example', 'src', 'src/dest',
                 aws_access_key_id='id',
                 aws_secret_access_key='key')



"""Register and confirm new build uploads with the LTD Keeper API.
"""

__all__ = ('register_build', 'confirm_build')

import logging
from urllib.parse import urljoin

import requests
import uritemplate

from .exceptions import KeeperError


def register_build(host, keeper_token, product, git_refs, dirnames=None):
    """Register a new build for a product on LSST the Docs.

    Wraps the ``POST /products/{product}/builds/`` **v2 API** endpoint.

    Parameters
    ----------
    host : `str`
        Hostname of LTD Keeper API server.
    keeper_token : `str`
        Auth token (`ltdconveyor.keeper.get_keeper_token`).
    product : `str`
        Name of the product in the LTD Keeper service.
    git_refs : `list` of `str`
        List of Git refs that correspond to the version of the build. Git refs
        can be tags or branches.
    dirnames : `list` of `str`
        A list of relative directory names in the site. Each directory will
        get its own presigned POST URL in the response from the LTD Keeper API.

    Returns
    -------
    build_info : `dict`
        LTD Keeper build resource.

    Raises
    ------
    ltdconveyor.keeper.KeeperError
        Raised if there is an error communicating with the LTD Keeper API.
    """
    logger = logging.getLogger(__name__)

    data = {
        'git_refs': git_refs
    }
    if dirnames is not None:
        data['directories'] = list(dirnames)

    endpoint_url = uritemplate.expand(
        urljoin(host, '/products/{p}/builds/'),
        p=product)

    r = requests.post(
        endpoint_url,
        auth=(keeper_token, ''),
        json=data,
        headers={'Accept': 'application/vnd.ltdkeeper.v2+json'}
    )

    if r.status_code != 201:
        raise KeeperError(r.json())
    build_info = r.json()
    logger.debug(
        'Registered a build for product %s:\n%s',
        product,
        build_info)
    return build_info


def confirm_build(build_url, keeper_token):
    """Confirm a build upload is complete.

    Wraps ``PATCH /builds/{build}``.

    Parameters
    ----------
    build_url : `str`
        URL of the build resource. Given a build resource, this URL is
        available from the ``self_url`` field.
    keeper_token : `str`
        Auth token (`ltdconveyor.keeper.get_keeper_token`).

    Raises
    ------
    ltdconveyor.keeper.KeeperError
        Raised if there is an error communicating with the LTD Keeper API.
    """
    data = {
        'uploaded': True
    }

    r = requests.patch(
        build_url,
        auth=(keeper_token, ''),
        json=data)
    if r.status_code != 200:
        raise KeeperError(r)



"""Tests for ltdconveyor.s3upload.upload_dir

These tests required the following environment variables:

``LTD_TEST_AWS_ID``
      AWS access key ID

``LTD_TEST_AWS_SECRET``
      AWS secret access key

``LTD_TEST_BUCKET``
      Name of an S3 bucket that already exists and can be used for testing.

The tests will be skipped if they are not available.

Note that this test will create a random uuid4) directory at the root of
``LTD_TEST_BUCKET``, though the test harness will attempt to delete it.
"""

import os
import shutil
import tempfile
import uuid
import logging
import mimetypes

import pytest
import boto3
import requests
from ltdconveyor.s3 import upload_dir


log = logging.getLogger(__name__)


@pytest.mark.skipif(os.getenv('LTD_TEST_AWS_ID') is None or
                    os.getenv('LTD_TEST_AWS_SECRET') is None or
                    os.getenv('LTD_TEST_BUCKET') is None,
                    reason='Set LTD_TEST_AWS_ID, '
                           'LTD_TEST_AWS_SECRET and '
                           'LTD_TEST_BUCKET')
def test_upload_dir(request):
    """Integration test of upload_dir and synchronnization upon file deletes.

    This is a test that operates more as an integrations test than a unit
    test.
    """
    bucket_name = os.getenv('LTD_TEST_BUCKET')
    aws_credentials = {
        'aws_access_key_id': os.getenv('LTD_TEST_AWS_ID'),
        'aws_secret_access_key': os.getenv('LTD_TEST_AWS_SECRET')}
    session = boto3.session.Session(**aws_credentials)

    logging.basicConfig(level=logging.ERROR)

    logging.getLogger('ltdconveyor.s3upload').level = logging.INFO
    logging.getLogger(__name__).level = logging.INFO

    paths = [
        'file1.txt',
        'file2.txt',
        'dir1/file11.txt',
        'dir1/file12.txt',
        'dir1/dir11/file111.txt',
        'dir1/dir11/file112.txt',
        'dir2/file21.txt',
        'dir2/file22.txt',
    ]

    temp_dir = tempfile.mkdtemp()
    temp_bucket_dir = str(uuid.uuid4())

    def cleanup():
        print("Cleaning up the bucket")
        _clean_bucket(session, bucket_name, temp_bucket_dir)
    request.addfinalizer(cleanup)

    _create_test_files(temp_dir, paths)

    surrogate_key = 'test-surrogate-key'
    cache_control_max_age = 3600
    upload_dir(bucket_name,
               temp_bucket_dir,
               temp_dir,
               surrogate_key=surrogate_key,
               upload_dir_redirect_objects=True,
               cache_control='max-age={0:d}'.format(cache_control_max_age),
               **aws_credentials)

    _test_objects_exist(session, bucket_name, temp_bucket_dir, paths)
    _test_content_types(session, bucket_name, temp_bucket_dir)
    _test_directory_redirects(session, bucket_name, temp_bucket_dir, paths)

    expected_headers = {
        'x-amz-meta-surrogate-key': surrogate_key,
        'Cache-Control': 'max-age={0:d}'.format(cache_control_max_age)}
    _test_headers(session, bucket_name, temp_bucket_dir, expected_headers)

    # Remove some files (on filesystem and in path manifest)
    shutil.rmtree(os.path.join(temp_dir, 'dir1'))
    paths.remove('dir1/file11.txt')
    paths.remove('dir1/file12.txt')
    paths.remove('dir1/dir11/file111.txt')
    paths.remove('dir1/dir11/file112.txt')
    os.remove(os.path.join(temp_dir, 'file2.txt'))
    paths.remove('file2.txt')

    upload_dir(bucket_name,
               temp_bucket_dir,
               temp_dir,
               **aws_credentials)
    _test_objects_exist(session, bucket_name, temp_bucket_dir, paths)

    shutil.rmtree(temp_dir)


def _create_test_files(temp_dir, file_list):
    for path in file_list:
        _write_file(temp_dir, path)


def _write_file(root_dir, rel_path):
    filepath = os.path.join(root_dir, rel_path)
    try:
        os.makedirs(os.path.dirname(filepath))
    except OSError:
        # probably exists; pickup real error on write
        pass

    with open(filepath, 'w') as f:
        f.write('Content of {0}'.format(os.path.basename(filepath)))


def _test_objects_exist(session, bucket_name, bucket_root, file_list):
    s3 = session.resource('s3')
    bucket = s3.Bucket(bucket_name)

    # dirnames = _file_list_dirnames(file_list, bucket_root)

    bucket_objects = []
    for obj in bucket.objects.filter(Prefix=bucket_root):
        # skip directory objects
        if not os.path.splitext(obj.key)[-1]:
            continue
        bucket_objects.append(obj.key)

    assert len(bucket_objects) == len(file_list)
    for path in file_list:
        found = False
        for bucket_path in bucket_objects:
            if bucket_path.endswith(path):
                found = True
                continue
        if not found:
            log.error('{0} not found in bucket'.format(bucket_path))
            assert False


def _test_headers(session, bucket_name, bucket_root, expected_headers):
    """Generically test that header key-value pairs in `expected_headers`
    actually are served by S3.
    """
    s3 = session.resource('s3')
    bucket = s3.Bucket(bucket_name)

    # see http://stackoverflow.com/a/34698521 for making object URLs
    bucket_location = s3.meta.client.get_bucket_location(Bucket=bucket_name)

    for obj in bucket.objects.filter(Prefix=bucket_root):
        object_url = "https://s3-{0}.amazonaws.com/{1}/{2}".format(
            bucket_location['LocationConstraint'],
            bucket_name,
            obj.key)
        r = requests.head(object_url)
        # skip directory redirect objects
        if 'x-amz-ltd-redirect' in r.headers:
            continue
        for key, expected_value in expected_headers.items():
            assert r.headers[key] == expected_value


def _test_content_types(session, bucket_name, bucket_root):
    """Verify that the expected Content-Type header was set."""
    s3 = session.resource('s3')
    bucket = s3.Bucket(bucket_name)

    # AWS api doesn't give Content-Type header, so we'll test it directly
    # via HTTP
    # see http://stackoverflow.com/a/34698521 for making object URLs
    bucket_location = s3.meta.client.get_bucket_location(Bucket=bucket_name)

    for obj in bucket.objects.filter(Prefix=bucket_root):
        # skip directory redirect objects
        if 'ltd-redirect' in obj.Object().metadata:
            continue
        guess, _ = mimetypes.guess_type(obj.key)
        object_url = "https://s3-{0}.amazonaws.com/{1}/{2}".format(
            bucket_location['LocationConstraint'],
            bucket_name,
            obj.key)
        if guess is not None:
            r = requests.head(object_url)
            assert r.headers['content-type'] == guess


def _file_list_dirnames(file_list, bucket_root):
    dirnames = [bucket_root]
    for path in file_list:
        _d = os.path.dirname(path)
        if len(_d) > 0:
            dirnames.append(os.path.join(bucket_root, _d))
    dirnames = list(set(dirnames))
    return dirnames


def _test_directory_redirects(session, bucket_name, bucket_root, file_list):
    """Verify that the directory redirect objects exist."""
    # Make a list of all directories, including the root directory
    dirnames = _file_list_dirnames(file_list, bucket_root)

    # see http://stackoverflow.com/a/34698521 for making object URLs
    s3 = session.resource('s3')
    # bucket = s3.Bucket(bucket_name)
    bucket_location = s3.meta.client.get_bucket_location(Bucket=bucket_name)

    for dirname in dirnames:
        # Try to request the object
        obj = s3.Object(bucket_name, dirname)
        object_url = "https://s3-{0}.amazonaws.com/{1}/{2}".format(
            bucket_location['LocationConstraint'],
            bucket_name,
            obj.key)
        r = requests.head(object_url)
        assert r.headers['x-amz-meta-dir-redirect'] == 'true'


def _clean_bucket(session, bucket_name, root_path):
    s3 = session.resource('s3')
    bucket = s3.Bucket(bucket_name)

    # Normalize directory path for searching patch prefixes of objects
    if not root_path.endswith('/'):
        root_path += '/'

    key_objects = [{'Key': obj.key}
                   for obj in bucket.objects.filter(Prefix=root_path)]
    if len(key_objects) == 0:
        return
    delete_keys = {'Objects': []}
    delete_keys['Objects'] = key_objects
    # based on http://stackoverflow.com/a/34888103
    s3.meta.client.delete_objects(Bucket=bucket.name,
                                  Delete=delete_keys)



"""Tests for ltdconveyor.cli.upload.
"""

import click
import pytest

from ltdconveyor.cli.upload import (_should_skip_travis_event,
                                    _get_git_refs,
                                    _get_travis_git_refs,
                                    _get_gh_actions_git_refs)


@pytest.mark.parametrize(
    'event_type,on_push,on_pr,on_api,on_cron,expected',
    [
        ('push', True, True, True, True, False),
        ('push', False, True, True, True, True),
        ('pull_request', True, True, True, True, False),
        ('pull_request', True, False, True, True, True),
        ('api', True, True, True, True, False),
        ('api', True, True, False, True, True),
        ('cron', True, True, True, True, False),
        ('cron', True, True, True, False, True),
    ])
def test_should_skip_travis_event(monkeypatch, event_type, on_push, on_pr,
                                  on_api, on_cron, expected):
    """Test _should_skip_travis_event with different combinations of
    user settings and TRAVIS_EVENT_TYPE variables.
    """
    import os

    def mock_env_var(_):
        return event_type

    monkeypatch.setattr(os, 'getenv', mock_env_var)
    assert _should_skip_travis_event(on_push, on_pr, on_api, on_cron) \
        is expected


def test_get_travis_git_refs(monkeypatch):
    import os

    monkeypatch.setattr(os, 'getenv', lambda _: 'my_branch')

    assert _get_travis_git_refs() == ['my_branch']


def test_get_travis_git_refs_no_env_var(monkeypatch):
    import os

    monkeypatch.setattr(os, 'getenv', lambda _: None)

    with pytest.raises(click.UsageError):
        _get_travis_git_refs()


def test_get_gh_actions_git_refs(monkeypatch):
    import os

    monkeypatch.setattr(os, 'getenv', lambda _: 'refs/heads/my_branch')

    assert _get_gh_actions_git_refs() == ['my_branch']


def test_get_gh_actions_git_refs_no_env_var(monkeypatch):
    import os

    monkeypatch.setattr(os, 'getenv', lambda _: None)

    with pytest.raises(click.UsageError):
        _get_gh_actions_git_refs()


@pytest.mark.parametrize(
    'env_var,ci_env,user_git_ref,expected',
    [
        # using travis branch
        ('my_branch', 'travis', None, ['my_branch']),
        # overriding travis on command line
        ('my_branch', 'travis', 'user-branch', ['user-branch']),
        # using GitHub actions branch
        ('refs/heads/my_branch', 'gh', None, ['my_branch']),
        ('refs/tags/my_tag', 'gh', None, ['my_tag']),
        ('refs/pull/my_pr', 'gh', None, ['my_pr']),
        # overriding GitHub actions on command line
        ('refs/heads/my_branch', 'gh', 'user-branch', ['user-branch']),
        # only using command line arg
        ('my_branch', None, 'user-branch', ['user-branch']),
        (None, None, 'user-branch', ['user-branch']),
        # parsing user git ref list
        (None, None, 'a b   c', ['a', 'b', 'c']),
    ]
)
def test_get_git_refs(monkeypatch, env_var, ci_env, user_git_ref,
                      expected):
    import os

    monkeypatch.setattr(os, 'getenv', lambda _: env_var)

    assert _get_git_refs(ci_env, user_git_ref) == expected



from .copy import copy_dir
from .delete import delete_dir
from .exceptions import S3Error
from .upload import (upload_dir, upload_file, upload_object,
                     create_dir_redirect_object, ObjectManager)
from .utils import open_bucket



"""Delete an S3 directory.
"""

__all__ = ('delete_dir',)

import logging
import boto3

from .exceptions import S3Error


def delete_dir(bucket_name, root_path,
               aws_access_key_id=None, aws_secret_access_key=None,
               aws_profile=None):
    """Delete all objects in the S3 bucket named ``bucket_name`` that are
    found in the ``root_path`` directory.

    Parameters
    ----------
    bucket_name : `str`
        Name of an S3 bucket.
    root_path : `str`
        Directory in the S3 bucket that will be deleted.
    aws_access_key_id : `str`
        The access key for your AWS account. Also set
        ``aws_secret_access_key``.
    aws_secret_access_key : `str`
        The secret key for your AWS account.
    aws_profile : `str`, optional
        Name of AWS profile in :file:`~/.aws/credentials`. Use this instead
        of ``aws_access_key_id`` and ``aws_secret_access_key`` for file-based
        credentials.

    Raises
    ------
    ltdconveyor.s3.S3Error
        Thrown by any unexpected faults from the S3 API.
    """
    logger = logging.getLogger(__name__)

    session = boto3.session.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key)
    s3 = session.resource('s3')
    client = s3.meta.client

    # Normalize directory path for searching patch prefixes of objects
    if not root_path.endswith('/'):
        root_path.rstrip('/')

    paginator = client.get_paginator('list_objects_v2')
    pages = paginator.paginate(Bucket=bucket_name, Prefix=root_path)

    keys = dict(Objects=[])
    for item in pages.search('Contents'):
        try:
            keys['Objects'].append({'Key': item['Key']})
        except TypeError:  # item is None; nothing to delete
            continue
        # Delete immediately when 1000 objects are listed
        # the delete_objects method can only take a maximum of 1000 keys
        if len(keys['Objects']) >= 1000:
            try:
                client.delete_objects(Bucket=bucket_name, Delete=keys)
            except Exception:
                message = 'Error deleting objects from %r' % root_path
                logger.exception(message)
                raise S3Error(message)
            keys = dict(Objects=[])

    # Delete remaining keys
    if len(keys['Objects']) > 0:
        try:
            client.delete_objects(Bucket=bucket_name, Delete=keys)
        except Exception:
            message = 'Error deleting objects from %r' % root_path
            logger.exception(message)
            raise S3Error(message)



"""S3 Upload client using presigned POST URLs generated by the LTD Keeper
server application.

For more information about S3 presigned POST URLs, see
https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-presigned-urls.html#generating-a-presigned-url-to-upload-a-file
"""

__all__ = ('prescan_directory', 'upload_dir', 'upload_file',
           'upload_directory_objects')

from copy import deepcopy
import mimetypes
import logging

import requests

from .exceptions import S3Error
from ..exceptions import ConveyorError


def prescan_directory(base_dir, _current_dir=None):
    """Make a list of all directories in a site, including the root
    directory.

    This function is used to the LTD Keeper API what presigned POST URLs to
    create for a build (see `ltdconveyor.keeper.register_build`).

    Parmameters
    -----------
    base_dir : `pathlib.Path`
        The (local) root directory of a web site.

    Returns
    -------
    dirnames : `list`
        A list of directory names, relative to the root directory. The root
        directory is represented by ``"/"``. All directory names end with
        ``"/"``.
    """
    dirs = []
    if _current_dir is None:
        _current_dir = base_dir
    dirs.append(format_relative_dirname(_current_dir, base_dir))
    for path in _current_dir.iterdir():
        if path.is_dir():
            dirs.extend(prescan_directory(base_dir, _current_dir=path))
    return dirs


def format_relative_dirname(directory, base_directory):
    """Formats a relative directory path in a way that's compatible with the
    presigned POST URLs.

    Parameters
    ----------
    directory : `pathlib.Path`
        The directory to compute a relative path/name for.
    base_directory : `pathlib.Path`
        The base directory.

    Returns
    -------
    name : `str`
        The relative directory name.

        Examples:

        - ``"base/`` relative to ``"/base/"`` is ``"/"``.
        - ``"base/a/`` relative to ``"/base/"`` is ``"a/"``.
        - ``"base/a/b`` relative to ``"/base/"`` is ``"a/b/"``.
    """
    name = str(directory.relative_to(base_directory))
    if name == '.':
        return '/'
    elif not name.endswith('/'):
        return name + '/'
    else:
        return name


def upload_dir(*, post_urls, base_dir, _current_dir=None):
    """Upload a local directory of files to S3 for an LSST the Docs build.

    Parameters
    ----------
    post_urls : `dict`
        This dictionary is obtained from the ``"post_prefix_urls"`` field of
        the `ltdconveyor.keeper.register_build` function. It contains
        presigned post POST URLs and fields for each directory in the site
        being uploaded (see the ``dirnames`` parameter to
        `~ltdconveyor.keeper.register_build`).
    base_dir : `pathlib.Path`
        Base directory of the site.

    See also
    --------
    upload_directory_objects
    """
    logger = logging.getLogger(__name__)

    if _current_dir is None:
        _current_dir = base_dir

    relative_dir = format_relative_dirname(_current_dir, base_dir)
    try:
        post_url = post_urls[relative_dir]
    except KeyError:
        logger.exception('A presigned POST URL is not available for the '
                         '%s directory', relative_dir)
        raise ConveyorError

    for path in _current_dir.iterdir():
        if path.is_file():
            upload_file(
                local_path=path,
                post_url=post_url['url'],
                post_fields=post_url['fields'])
        elif path.is_dir():
            upload_dir(
                post_urls=post_urls,
                base_dir=base_dir,
                _current_dir=path)


def upload_file(*, local_path, post_url, post_fields):
    """Upload a file using a presigned POST URL to S3.

    This function should primarily be used by `upload_dir`.

    Parameters
    ----------
    local_path : `pathlib.Path`
        Path to the file being uploaded.
    post_url : `str`
        URL to upload to.
    post_fields : `dict`
        Dictionary of fields for the POST. Generally these fields are created
        by the LST Keeper API server when it generates the presigned URLs.

    Raises
    ------
    ltdconveyor.s3.exceptions.S3Error
        Raised if the upload fails.
    """
    logger = logging.getLogger(__name__)

    filename = local_path.name
    # don't want to globally change post_fields
    post_fields = deepcopy(post_fields)

    # Detect the Content-Type. This is a required field for the post URL.
    content_type, content_encoding = mimetypes.guess_type(filename,
                                                          strict=False)
    if content_type is not None:
        post_fields['Content-Type'] = content_type
    else:
        post_fields['Content-Type'] = 'application/octet-stream'

    with open(local_path, 'rb') as f:
        files = {'file': (filename, f)}
        http_response = requests.post(post_url, data=post_fields, files=files)
    if http_response.status_code == 204:
        logger.debug('Uploaded %s using presigned POST URL fields %s',
                     local_path, post_fields)
    else:
        logger.error('Error uploading %s (code %i) using presigned POST URL '
                     'fields %s', local_path, http_response.status_code,
                     post_fields)
        raise S3Error


def upload_directory_objects(*, post_urls):
    """Upload directory redirect objects for an LSST the Docs product build.

    Parameters
    ----------
    post_urls : `dict`
        This dictionary is obtained from the ``"post_dir_urls"`` field of
        the `ltdconveyor.keeper.register_build` function. It contains
        presigned post POST URLs and fields for each directory in the site
        being uploaded (see the ``dirnames`` parameter to
        `~ltdconveyor.keeper.register_build`).

    See also
    --------
    upload_dir

    Notes
    -----
    Directory redirect objects in S3 are named after directories, don't have
    a trailing slash in their key name, and don't have any content. What
    they *do* have is a ``x-amz-meta-dir-redirect`` header key. The LSST the
    Docs Fastly configuration looks for this header, and when detected,
    redirects the request to the associated ``*/index.html`` object.
    """
    logger = logging.getLogger(__name__)
    for dirname, post_url in post_urls.items():
        url = post_url['url']
        fields = post_url['fields']
        files = {'file': ('', '')}
        http_response = requests.post(url, data=fields, files=files)
        if http_response.status_code == 204:
            logger.debug(
                'Uploaded directory object for %s',
                dirname)
        else:
            logger.error(
                'Error uploading directory object for %s (code %i) using '
                'presigned POST URL fields %s', dirname,
                http_response.status_code, fields)
            raise S3Error



__all__ = ('s3', 'fastly', 'ConveyorError')

import logging

from pkg_resources import get_distribution, DistributionNotFound

from . import s3
from . import fastly
from .exceptions import ConveyorError

# Allow applications to control the logging handler
logging.getLogger(__name__).addHandler(logging.NullHandler())

try:
    __version__ = get_distribution('ltd-conveyor').version
except DistributionNotFound:
    # Package is not installed
    __version__ = 'unknown'
